/*
 * arp		This file contains an implementation of the command
 *		that maintains the kernel's ARP cache.  It is derived
 *		from Berkeley UNIX arp(8), but cleaner and with sup-
 *		port for devices other than Ethernet.
 *
 * NET-TOOLS	A collection of programs that form the base set of the
 *		NET-3 Networking Distribution for the LINUX operating
 *		system.
 *
 * Version:	arp 1.83 (1998-02-13)
 *
 * Maintainer:	Bernd 'eckes' Eckenfels, <net-tools@lina.inka.de>
 *
 * Author: 	Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *
 * Changes:
 *		(based on work from Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>)
 *              Alan Cox        :	modified for NET3
 *              Andrew Tridgell :	proxy arp netmasks
 *              Bernd Eckenfels :	-n option
 *              Bernd Eckenfels :	Use only /proc for display
 *	 {1.60}	Bernd Eckenfels :	new arpcode (-i) for 1.3.42 but works 
 *					with 1.2.x, too
 *	 {1.61}	Bernd Eckenfels :	more verbose messages
 *	 {1.62}	Bernd Eckenfels :	check -t for hw adresses and try to
 *					explain EINVAL (jeff)
 *970125 {1.63}	Bernd Eckenfels :	-a print hardwarename instead of tiltle
 *970201 {1.64} Bernd Eckenfels :	net-features.h support
 *970203 {1.65} Bernd Eckenfels :	"#define" in "#if", 
 *					-H|-A additional to -t|-p
 *970214 {1.66} Bernd Eckenfels :	Fix optarg required for -H and -A
 *970412 {1.67} Bernd Eckenfels :	device=""; is default
 *970514 {1.68}	Bernd Eckenfels :	-N and -D
 *970517 {1.69}	Bernd Eckenfels :	usage() fixed
 *970622 {1.70} Bernd Eckenfels :	arp -d priv
 *970106 {1.80} Bernd Eckenfels :	new syntax without -D and with "dev <If>",
 *					ATF_MAGIC, ATF_DONTPUB support. 
 *					Typo fix (Debian Bug#5728 Giuliano Procida)
 *970803 {1.81} Bernd Eckenfels :	removed junk comment line 1
 *970925 {1.82} Bernd Eckenfels	:	include fix for libc6
 *980213 (1.83) Phil Blundell:		set ATF_COM on new entries
 *980629 (1.84) Arnaldo Carvalho de Melo: gettext instead of catgets
 *050614 Dave Stuart (linux@justdave.us)Removed all code not used for options a or i.
 *
 *
 *		This program is free software; you can redistribute it
 *		and/or  modify it under  the terms of  the GNU General
 *		Public  License as  published  by  the  Free  Software
 *		Foundation;  either  version 2 of the License, or  (at
 *		your option) any later version.
 *
 * 
 * 6/22/05
 * Customized specifically for FREESCO by Lewis Baughman.
 * Many functions removed and a custome output format.
 *
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
/* #include <linux/netdevice.h> */
/* #include <linux/if_arp.h>    */
#include <net/if_arp.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include "net-support.h"
#include "pathnames.h"
/*
#include "version.h"
*/
#include "config.h"
#include "intl.h"

#define DFLT_AF	"inet"
#define DFLT_HW	"ether"

#define FEATURE_ARP
/* #include "lib/net-features.h"*/
/*
char *Release = RELEASE,
     *Version = "arp 1.84f";
*/
int what = 0;

int opt_n = 0;				/* do not resolve addresses	*/
int opt_N = 0;				/* use symbolic names           */
int opt_D = 0;				/* HW-address is devicename     */
int opt_e = 0;				/* 0=BSD output, 1=new linux    */
int opt_a = 0;				/* all entries, substring match */
struct aftype *ap;			/* current address family	*/
struct hwtype *hw;			/* current hardware type	*/
int sockfd=0;				/* active socket descriptor     */
int hw_set = 0;				/* flag if hw-type was set (-H)	*/
char device[16]="";			/* current device		*/
static void usage(void);


/* Print the contents of an ARP request block. */
static void
arp_disp(char *ip, int arp_flags,char *hwa)
{
  printf("%s,", ip);

  if (!(arp_flags & ATF_COM)) {
	if (arp_flags & ATF_PUBL)
		printf("*\n");
	else
		printf("<incomplete>\n");
  } else {
	printf("%s\n", hwa);
  }
}

/* Print the contents of an ARP request block. */
static void
arp_disp_2(char *name, char *ip, int type,int arp_flags,char *hwa,char *mask,char *dev)
{
  struct hwtype *xhw;
  
  xhw = get_hwntype(type);
  if (xhw == NULL) 
    xhw = get_hwtype(DFLT_HW);
    


  printf("%s (%s) at ", name, ip);

  if (!(arp_flags & ATF_COM)) {
	if (arp_flags & ATF_PUBL)
		printf("* ");
	else
		printf("<incomplete> ");
  } else {
	printf("%s [%s] ", hwa, xhw->name);
  }
  
  if (arp_flags & ATF_NETMASK)
	printf("netmask %s ", mask);
  
  if (arp_flags & ATF_PERM) printf("PERM ");
  if (arp_flags & ATF_PUBL) printf("PUP ");
#ifdef HAVE_ATF_MAGIC
  if (arp_flags & ATF_MAGIC) printf("AUTO ");
#endif
#ifdef HAVE_ATF_DONTPUB
  if (arp_flags & ATF_DONTPUB) printf("DONTPUB ");
#endif
  if (arp_flags & ATF_USETRAILERS) printf("TRAIL ");

  printf("on %s\n", dev);
}


/* Display the contents of the ARP cache in the kernel. */
static int
arp_show(char *name)
{
  char host[100];
  struct sockaddr sa;
  char ip[100];
  char hwa[100];
  char mask[100];
  char line[200];
  char dev[100];
  int type,flags;
  FILE *fp;
  char *hostname;
  int num,entries=0,showed=0;

  host[0]='\0';
  
  if (name != NULL) {
  	/* Resolve the host name. */
	  host[(sizeof host)-1] = 0;
  	strncpy(host, name, (sizeof host)-1);
  	if (ap->input(0, host, &sa) < 0) {
		ap->herror(host);
		return(-1);
  	}
  	strcpy(host,ap->sprint(&sa, 1));
  }
  
  /* Open the PROCps kernel table. */
  if ((fp = fopen(_PATH_PROCNET_ARP, "r")) == NULL) {
	perror(_PATH_PROCNET_ARP);
	return(-1);
  }

  /* Bypass header -- read until newline */
  if (fgets(line, sizeof(line), fp) != (char *)NULL) {
	strcpy(mask,"-");
	strcpy(dev,"-");
	/* Read the ARP cache entries. */
	for(;fgets(line,sizeof(line),fp);)
	{
		num=sscanf(line,"%s 0x%x 0x%x %100s %100s %100s\n",
		                               ip,&type,&flags,hwa,mask,dev);
		if(num<4)
			break;
		
		entries++;	
		/* if the user specified hw-type differs, skip it */
		if (hw_set && (type != hw->type))
			continue;
			
		/* if the user specified address differs, skip it */
		if (host[0] && strcmp(ip,host))
			continue;
		
		/* if the user specified device differs, skip it */
		if (device[0] && strcmp(dev,device))
			continue;

		showed++;
		/* This IS ugly but it works -be */
		if (opt_n)
			hostname="?";
		else {
			if (ap->input(0, ip,&sa) < 0)
				hostname=ip;
			else
				hostname = ap->sprint(&sa, opt_n | 0x8000);
			if (strcmp(hostname, ip)==0)
				hostname="?";
		}

		if (opt_e)
			arp_disp_2(hostname,ip,type,flags,hwa,mask,dev);
		else
			arp_disp(ip,flags,hwa);
	}
  }

  (void) fclose(fp);
  return(0);
}
/*
static void
version(void)
{
  fprintf(stderr, "%s\n%s\n",Release,Version);
  exit(-1);
}
*/
static void
usage(void)
{
  fprintf(stderr, _("Usage: arpf [-i if] [-s]\n"));
  exit(-1);
}

int
main(int argc, char **argv)
{

  /* Initialize variables... */
  if ((hw = get_hwtype(DFLT_HW)) == NULL) {
	return(-1);
  }
  if ((ap = get_aftype(DFLT_AF)) == NULL) {
	return(-1);
  }

  /* Fetch the command-line arguments. */
  /* opterr = 0; */
  while ((i = getopt_long(argc, argv, "si:h?")) != EOF) switch(i) {
	case 's':
		what = 1;
		opt_a = 1;
		break;

	case 'i':
		strncpy(device,optarg,sizeof(device)-1);
		device[sizeof(device)-1]='\0';
		break;

	case 'h':
	default:
		usage();
  }

  if ((sockfd = socket(AF_INET,SOCK_DGRAM,0)) <0)
  {
  	perror("socket");
  	exit(-1);
  }

  /* Now see what we have to do here... */
  switch(what) {
	case 0:
		opt_e = 1;
		what = arp_show(argv[optind]);
		break;
		
	case 1:		/* show an ARP entry in the cache */
		what = arp_show(argv[optind]);
		break;

	default:
		usage();
  }

  exit(what);
}
