/*  
    A logrotate program for FREESCO written to Lewis's specifications.
    Written by Dave Stuart <freesco@justdave.us> 12-Dec-2005

    06-Jan-06  Added code to display /proc/net/dev on tty5.
*/


#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>

// There are only 1440 minutes in a day so this is the max that could be defined.

int restrict_times[1440];
int restrict_count = 0;
int loop_count;
int screen_row = 0;
FILE *file, *file2;
char temp[5000], *value;

void get_restrict_times() 
   {
   file = fopen ("/etc/restrict.cfg", "r");
   if (file != NULL) {
      restrict_count = 0;
      while (fgets (temp, 5000, file) != 0) {
         // Ignore comment lines
         if (temp[0] != '#') {
            // Ignore the first 2 values because the times are in fields 3 and 4.(if they exist at all)
            value = strtok (temp, ",");
            value = strtok (NULL, ",");
            value = strtok (NULL, ",");
            // Convert times to minutes from midnight
            if (value != NULL)
               restrict_times[restrict_count++] = (atoi (value) / 100) * 60 + (atoi (value) % 100);
            value = strtok (NULL, ",");
            if (value != NULL)
               restrict_times[restrict_count++] = (atoi (value) / 100) * 60 + (atoi (value) % 100);
            }
         }
      fclose (file);
      fclose (file2);
      }
   return;
   }

int main (int argc, char *argv[])
{
   struct stat filestats;
   int running_from_floppy = 0, logsize = 0, secure_logsize = 0, 
   enable_logger = 0, i;
   int current_time;

   struct tm *time_structure;
   time_t leases, passwd, restrict, timedate, htpasswd;
 
 // Get the command line arguments
   if ((strcmp (argv[1], "fd") == 0) || (strcmp (argv[1], "ram1") == 0))
      running_from_floppy = 1;
   logsize  = atoi (argv[2]);
   secure_logsize = atoi (argv[3]); 
   if (strcmp (argv[4], "y") == 0)
      enable_logger = 1;
   
// Set up the current modify times for the files.
 
   stat ("/etc/dhcpd.leases", &filestats);
   leases = filestats.st_mtime;
   
   stat ("/etc/passwd", &filestats);
   passwd = filestats.st_mtime;
   
   stat ("/wwa/cgi/.htpasswd", &filestats);
   htpasswd = filestats.st_mtime;

   stat ("/etc/restrict.cfg", &filestats);
   restrict = filestats.st_mtime;
   get_restrict_times();

// make it a lower priority
   nice(10);
      
// Now watch for any changes
   while (1) { 

// The dhcpd.leases is the only file not saved on a floppy install
      if (!running_from_floppy) {
         stat("/etc/dhcpd.leases", &filestats);
         if (leases != filestats.st_mtime) {
           system ("cat /etc/dhcpd.leases >/boot/etc/dhcpd.liz");
           system ("sync");
           leases = filestats.st_mtime;
           }
      }
      stat ("/etc/passwd", &filestats);
      if (passwd != filestats.st_mtime) {
         system ("cp /etc/passwd /boot/etc/passwd");
         system ("sync");
         passwd = filestats.st_mtime;
         }
         
      stat ("/wwa/cgi/.htpasswd", &filestats);
      if (htpasswd != filestats.st_mtime) {
         system ("cp /wwa/cgi/.htpasswd /boot/etc/.htpasswd");
         system ("sync");
         htpasswd = filestats.st_mtime;
         }
         
      stat ("/etc/restrict.cfg", &filestats);
      if (restrict != filestats.st_mtime) {
         restrict = filestats.st_mtime;
         file = fopen ("/etc/restrict.cfg", "r");
         if (file != NULL) {
             file2 = fopen ("/boot/etc/restrict.cfg", "w");
             while (fgets (temp, 5000, file) != 0) {
                // Ignore comment lines
                if (temp[0] != '#') {
                   fputs (temp, file2);
                }
            }
        fclose (file);
        fclose (file2);
         }
	 system ("sync");
         system ("fork rc_masq restart restrict.cfg");
      }
      
      if (restrict_count != 0) {
          // Get the number of seconds since January 1, 1970.
          timedate = time(NULL);
          // Format time into minutes from midnight
          time_structure = localtime (&timedate);
          current_time = time_structure->tm_hour * 60 + time_structure->tm_min;
          for (i = 0; i < restrict_count; i++)
             if (restrict_times[i] == current_time) {
                system ("fork rc_masq restart restrict.cfg");
                sleep (49);
                break;
             }
      }
     
// Check the log sizes
      stat ("/var/messages", &filestats);
      if (filestats.st_size > logsize) {
        system ("cat /var/messages >/var/messages.old");
        file = fopen ("/var/messages", "w");
        fclose (file);
	system ("savelogs /var/messages.old");
	system ("fork sendlogs messages /var/messages.old");
        system ("sync");
        }
        
      stat ("/var/secure", &filestats);
      if (filestats.st_size > secure_logsize) {
        system ("cat /var/secure >/var/secure.old");
        file = fopen ("/var/secure", "w");
        fclose (file);
	system ("savelogs secure.old");
	system ("fork sendlogs secure /var/secure.old");
        system ("sync");
     }

 /* Display /proc/net/dev on tty5 if it is enabled*/
     if (enable_logger) {
         file2 = fopen ("/dev/tty5", "w");
         fputs ("\e[2J", file2);
         for (loop_count = 0; loop_count < 12; loop_count++)   {
              file = fopen ("/proc/net/dev", "r");
              fputs ("\e[0;0H", file2);
              while (fgets (temp, 5000, file) != 0) {
                 if (strstr (temp, "dummy") == NULL) {
                     fputs (temp, file2);
                     fputs ("\r", file2);
                 }
              }
	      sleep (1);
              fclose (file);
         }
	 fclose (file2);
     } else {
         sleep (11);
     }
   }
// Should never get here.
   return (0);
}
// Compile with
// gcc -O2 logrotate.c -o logrotate
