/*
                          COPYRIGHT 1988
              Evans & Sutherland Computer Corporation
                        Salt Lake City, Utah
                        All Rights Reserved.

     THE INFORMATION  IN  THIS  SOFTWARE  IS  SUBJECT  TO  CHANGE
     WITHOUT  NOTICE  AND SHOULD NOT BE CONSTRUED AS A COMMITMENT
     BY  EVANS  &  SUTHERLAND.   EVANS  &  SUTHERLAND   MAKES  NO
     REPRESENTATIONS  ABOUT  THE SUITABILITY OF THIS SOFTWARE FOR
     ANY PURPOSE.  IT IS SUPPLIED  "AS  IS"  WITHOUT  EXPRESS  OR
     IMPLIED WARRANTY.

     IF THE SOFTWARE IS MODIFIED IN A MANNER CREATING  DERIVATIVE
     COPYRIGHT  RIGHTS,  APPROPRIATE LEGENDS MAY BE PLACED ON THE
     DERIVATIVE WORK IN ADDITION TO THAT SET FORTH ABOVE.

     Permission  to  use,  copy,  modify,  and  distribute   this
     software  and  its documentation for any purpose and without
     fee is hereby granted, provided  that  the  above  copyright
     notice  appear  in  all  copies  and that both the copyright
     notice and this permission notice appear in supporting docu-
     mentation,  and  that  the name of Evans & Sutherland not be
     used in advertising or publicity pertaining to  distribution
     of the software without specific, written prior permission.

Written by:

                        Robert C. Pendleton <bobp@hal.com>

Grateful acknowledgement is made of code and ideas contributed by

 	Ian Donaldson,
 	Department of Communications & Electronic Engineering,
 	Royal Melbourne Institute of Technology,
 	Melbourne, Australia.

$Header: ile.c,v 2.5 98/01/14 10:31:56 toehser Exp $
*/

/*
ile is compiled using:

cc ile.c -o ile -ltermcap
*/

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <pwd.h>
#include <utmp.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>


#ifdef INCLUDE_CURSES
#include <curses.h>
#include <term.h>
#endif

#ifdef USE_DIR
#include <sys/dir.h>
#define dirent direct
#endif

#ifdef USE_DIRENT
#include <dirent.h>
#endif

#include <unistd.h>
#include <stdlib.h>

/* This is for the option to list matching files when a double tab is typed */
#define DOUBLE_TAB

#ifndef MAXNAMLEN
#define MAXNAMLEN   255
#endif

#include "ile.h"

/*------------------------------------------------------------------*/
/*
Definitions of system stuff.
*/

extern int errno;

long lseek();
time_t time();

/*------------------------------------------------------------------*/

#define READ  0
#define WRITE 1
#define ERROR 2

#define BUFFER_SIZE 255

#define HISTORY_SIZE 101

#define USER_NAME_SIZE 8

#define EOL (-2)
/*------------------------------------------------------------------*/

/* special characters used by ile */

#define del '\177'

#define CA '\1'
#define CB '\2'
#define CC '\3'
#define CD '\4'
#define CE '\5'
#define CF '\6'
#define bel '\7'
#define bs '\10'
#define CI '\11'
#define nl '\12'
#define CK '\13'
#define CL '\14'
#define cr '\15'
#define CN '\16'
#define CO '\17'
#define CP '\20'
#define CQ '\21'
#define CR '\22'
#define CS '\23'
#define CT '\24'
#define CU '\25'
#define CV '\26'
#define CW '\27'
#define CX '\30'
#define CY '\31'
#define CZ '\32'

#define esc '\33'

/*------------------------------------------------------------------*/
/* areas and varaibles used to get termcap information */

char *getenv();

#ifndef INCLUDE_CURSES
char *tgetnum();
char *tgetflag();
char *tgetstr();
#endif

char termcap_entry[1024];	/* termcap entry for the users terminal */
char term_seqs[1024];		/* area to store control sequences in */
char *where = term_seqs;

char *cle;			/* move cursor left one space */

char *cce;			/* clear to end of line */

char *cbl;			/* audible bell */

char *cnl;			/* new line character */

char *ccr;			/* carriage return */

/*------------------------------------------------------------------*/
/*
Some various defines.
*/


#define MAX_FILES 5000
void query_path(unsigned char ch);
char defaultpath[255] = "";
char currentdir[MAXNAMLEN + 1] = "";
unsigned char last_char = ' ';
char listing[MAX_FILES][MAXNAMLEN + 1];
#ifdef DOUBLE_TAB
void show_files(unsigned char ch);
#endif /* DOUBLE_TAB */

/*------------------------------------------------------------------*/

/* tty status flags */
/*
  The original tty status flags are stored so that they can be
  restored when ile exits.
*/

int windowchanged;

/*------------------------------------------------------------------*/

/* file descriptors for tty and pty */

int master_pty;
int slave_tty;

/* the names of the tty and pty opened by getpty */

char ttydev[20];
char ptydev[20]; 

/* defaulkt ptty/tty name patterns */
#ifndef PTYDEV
#define PTYDEV          "/dev/ptyxx"
#endif

#ifndef TTYDEV
#define TTYDEV          "/dev/ttyxx"
#endif 

/* path and name of the lock file */

char lock[] = "/tmp/ile.lock";

/*------------------------------------------------------------------*/
/*
  getpty opens a pty, storing file descriptors in pty and tty.
  It trys pairs in order until it finds a pair that is not in use.
*/

#ifdef HAVE_MERGED_PTY

#define SCO_PTY         "/dev/ptyp"
#define SCO_TTY         "/dev/ttyp"
#define MAX_SCO_PTYS    256

#endif 

getpty(pty, tty)
    int *pty;
    int *tty;

{
    int devindex ;
    int letter;

    static char ptychar1[] = "pqrstuvwxyz";
    static char ptychar2[] = "0123456789abcdef";


#ifdef HAVE_MERGED_PTY
                        /* this may also be SYSV i386 specific - untested */
    struct utmp line;
    static int sco_devindex = 0;
    char sco_ttydev[20];
    char sco_ptydev[20];

    /*
     * First check to for sco merged ptys.
     */
    if (1)  /* mergedPtys */
    {
        for (; sco_devindex < MAX_SCO_PTYS; sco_devindex++) 
        {
	        sprintf (sco_ttydev, "%s%d", SCO_TTY, sco_devindex);
	        sprintf (sco_ptydev, "%s%d", SCO_PTY, sco_devindex);

	        if (access(sco_ttydev, F_OK) < 0)
	        {

		    sco_devindex = MAX_SCO_PTYS;
	    	    break;
	        }
    
                strcpy(line.ut_line,strrchr(sco_ptydev,'/')+1);
                setutent();

                if ((getutline(&line) == NULL) &&
	            (*pty = open (sco_ptydev, O_RDWR)) >= 0) 
	        {
	    	    /* 
		     * Set things up for our next entry
		     * into this function!
		     */
		    if ((*tty = open(sco_ttydev, O_RDWR)) < 0)
		    {
		        (void) close(*pty);
		        continue;
		    }
		    (void) sco_devindex++;
		    strcpy(ttydev, sco_ttydev);
		    strcpy(ptydev, sco_ptydev);
/*
printf("ptydev=[%s] fd=%d\n", ptydev, *pty);
printf("ttydev=[%s] fd=%d\n", ttydev, *tty );
*/
                    endutent();
	    	    return ;
	        }
	}
    }
#endif  /* HAVE_MERGED_PTY */

    strcpy(ptydev, PTYDEV);   /* Default format names - non SCO */
    strcpy(ttydev, TTYDEV);

    letter = 0;
    while (letter < 11)
    {
	ttydev[strlen(ttydev) - 2] = ptychar1[letter];
	ptydev[strlen(ptydev) - 2] = ptychar1[letter];
	letter++;

	devindex = 0;
	while (devindex < 16)
	{
	    ttydev[strlen(ttydev) - 1] = ptychar2[devindex];
	    ptydev[strlen(ptydev) - 1] = ptychar2[devindex];
	    devindex++;

	    if ((*pty = open(ptydev, O_RDWR)) >= 0)
	    {
		if ((*tty = open(ttydev, O_RDWR)) >= 0)
		{
		    return;
		}
		else
		{
		    (void) close(*pty);
		}
	    }
	}
    }

    (void) fprintf(stderr, "ile: unable to allocate pty/tty pair\n");
    exit(1);
    /* NOTREACHED */
}

/*------------------------------------------------------------------*/
/*
Termcap entries may have a sequences of digits optionally followed
by a '*' in front of the actual sequence. This routine increments
the pointer past this information.

Terminfo entries have %sequences which this rtn strips
*/
void
strip(ptr)
    char **ptr;
{
#ifdef OLD

    while (('0' <= **ptr) && (**ptr <= '9'))
    {
	(*ptr)++;
    }
 
    if (**ptr == '*')
    {
	(*ptr)++;
    }
#else

char *p = *ptr;
while (*p)
{
    if (*p == '%' )
    {
        char *p1 = p+1;
        char *d = p;
        if (*p1 == 'p' )
            p1 += 2;
        else
            p1++;    
        while (*p1)
        {
            *d++ = *p1++;
        }
        *d = '\0';
    }
    p++;
}

#endif
}
/*------------------------------------------------------------------*/
/*
Set up everything needed to use the control sequences from the
termcap entry for the terminal.
*/
void
get_termcap()
{
    char *terminal_type;	/* type of terminal */
    
    /* get the terminal name */

    terminal_type = getenv("TERM");

    /* get termcap entry */

    if (tgetent(termcap_entry, terminal_type) < 1)
    {
	(void) fprintf(stderr, "ile: can't find %s\n", terminal_type);
	exit(1);
	/* NOTREACHED */
    }

    /* get the control sequences ile needs */

    if ((cbl = tgetstr("bl", &where)) == NULL)
    {
	cbl = "\7";
    }

    if ((cnl = tgetstr("nl", &where)) == NULL)
    {
	cnl = "\n";
    }

    if ((ccr = tgetstr("cr", &where)) == NULL)
    {
	ccr = "\r";
    }

  /* cursor left */
    cle = tgetstr("le", &where);
    if ( cle  == NULL)
        cle = tgetstr("LE", &where);
    if ( cle == NULL )
        cle = tgetstr("cub", &where);
    if ( cle == NULL)
        cle = tgetstr("cub1", &where);
    if ( cle == NULL && tgetflag("bs") )
        cle = "\b";


    /* clr to EOL */
    cce = tgetstr("ce", &where);
    if (cce  == NULL )
        cce = tgetstr("el", &where);

    if (cle == NULL || cce == NULL)
    {
	(void) fprintf(stderr,
	    "ile: can't run on %s\n(need CsrLeft and ClrEOL capabilities \"le\" and \"ce\" or  \"cub1\" and \"el\")\n",
	    terminal_type);
	exit(1);
	/* NOTREACHED */
    }

    /* strip timing info from strings */

    strip(&cle);
    strip(&cce);
    strip(&cbl);
    strip(&cnl);
    strip(&ccr);
}

/*------------------------------------------------------------------*/
/*
 * set/clear the utmp slot for the pty
 */
int
setutmp(fd, set)
{
#ifdef HAVE_GETUT   /* getutline */

    struct utmp ut;
    char *cp;


    setutent();
    d_zero((char *) &ut, sizeof(ut));

    /* skip "/dev/" */

    cp = strrchr(ttydev, '/');
    if (cp == 0)
    {
        cp = ttydev;
    }
    else
    {
        cp++;
    }
    (void) strncpy(ut.ut_line, cp, sizeof(ut.ut_line));


    if (set)
    {
        struct passwd *pw;

	pw = getpwuid(getuid());
	if (pw == 0)
	{
	    (void) fprintf(stderr, "ile: who are you?\n");
	    return (-1);
	}

	(void) strncpy(ut.ut_user, pw->pw_name, sizeof(ut.ut_line));
        ut.ut_type = USER_PROCESS;
        ut.ut_pid = getpid();
        (void) time(&ut.ut_time);
    }
    else    /* clear entry */
    {
        if (getutline(&ut) == NULL )
        {
           /* perror("ile - getutline cannot position"); */
            return -1;
        }
        ut.ut_user[0] = '\0';
        ut.ut_type   = DEAD_PROCESS;
    }


    if (pututline(&ut) == NULL )
    {
        /* perror("ile - pututline cannot set"); */
        return -1;
    }
    return 0;
#endif    
#ifdef HAVE_TTYSLOT_UT

    int old0;
    int old1;
    int old2;
    int slot;
    int f;
    struct utmp ut;


    /* Must make fd's 0,1,2 correspond to slave_tty for ttyslot() to
     * function.  Ugh!  Why doesn't ttyslot() accept a fd argument?
     * 
     * save fd's */

    old0 = dup(0);
    old1 = dup(1);
    old2 = dup(2);

    if (old0 == -1 || old1 == -1 || old2 == -1)
    {
	perror("ile: dup");
	return (-1);		/* file table full ? */
    }

    /* set fd's 0,1,2 for ttyslot() */

    (void) dup2(fd, 0);
    (void) dup2(fd, 1);
    (void) dup2(fd, 2);

    slot = ttyslot();

    /* put the fd's back */

    (void) dup2(old0, 0);
    (void) dup2(old1, 1);
    (void) dup2(old2, 2);

    (void) close(old0);
    (void) close(old1);
    (void) close(old2);

    if (slot < 0)
    {
	(void) fprintf(stderr, "ile: don't know where you are\n");
	return (-1);
    }

    f = open("/etc/utmp", O_WRONLY);
    if (f == -1)
    {
	return (-1);
    }

    d_zero((char *) &ut, sizeof(ut));

    if (set)
    {
	struct passwd *pw;
	char *cp;

	pw = getpwuid(getuid());
	if (pw == 0)
	{
	    (void) fprintf(stderr, "ile: who are you?\n");
	    (void) close(f);
	    return (-1);
	}

	/* skip "/dev/" */

	cp = strrchr(ttydev, '/');
	if (cp == 0)
	{
	    cp = ttydev;
	}
	else
	{
	    cp++;
	}

	(void) strncpy(ut.ut_line, cp, sizeof(ut.ut_line));
	(void) strncpy(ut.ut_name, pw->pw_name, sizeof(ut.ut_line));
	(void) time(&ut.ut_time);
    }
    (void) lseek(f, (long) (slot * sizeof(struct utmp)), SEEK_SET);
    (void) write(f, (char *) &ut, sizeof(ut));
    (void) close(f);
    return (0);

#endif
}
/*------------------------------------------------------------------*/
/*
clean up and leave.

This function is bound to the SIGCHLD signal so that when the
child process exits, so does ile. It is also called when an exception
is detected by select() in ile().
*/
void
clean_up()
{

    /* kill off the child process */

    tty_killpg(slave_tty);

    /* restore terminal status */
    if ( tty_reset(READ) < 0 )
    {
        fprintf(stderr, "tty_reset() failed! Going sane....\n");
            (void) tty_sane(slave_tty);
    }

    /* "logout" the user */

    (void) setutmp(slave_tty, FALSE);

    /* clean up the tty/pty pair */

    (void) close(master_pty);
    (void) close(slave_tty);

    /* make things look nice */

    fputs(cnl, stdout);

    exit(0);
    /* NOTREACHED */
}
/*------------------------------------------------------------------*/
/*
Write a line to the slave_tty.
Get the slave_tty parameters, turn off echo, send the line to the
slave_tty, restore the slave_tty paramters to the way they were
before. If echo was already off, this will have no effect.
*/
write_line(line, length)
    char *line;
    int length;
{
    
    /* get the current parameters */

    tty_noecho( slave_tty, PUSH );

    (void) write(master_pty, line, length);

     tty_noecho( slave_tty, POP );


}

/*------------------------------------------------------------------*/
/*
The editing routines are called through the edit variable. This allows
the quote and escape commands to be implemented as a straight forward
state machine instead of requiring state flags and complex switch
statements.
*/
/*------------------------------------------------------------------*/

/* line edit buffer */

static char line[BUFFER_SIZE];

static int point;		/* insertion point */
static int length;		/* total chars in buffer */


/* procedure to edit next character */

void (*edit) ();

/* history buffer */

struct
{
    int length;
    char *line;
} hist[HISTORY_SIZE];

int head;			/* insertion point */
int here;			/* current displayed line */

/*------------------------------------------------------------------*/
/*
The delimiter vector is used by the forward, backward, and delete
word operations to decide that a character is a delimiter.
*/
/*------------------------------------------------------------------*/

#define CHAR_SET_SIZE 255
#define CHAR_MASK 0377

char delimit[CHAR_SET_SIZE];

/*------------------------------------------------------------------*/
/*
The action_table is used to bind sequences of keys to operations or strings.
*/
/*------------------------------------------------------------------*/

typedef enum
{
    is_action, is_string
} action_type;

struct
{
    action_type flag;
    union
    {
	void (*action) ();
	char *string;
    } aors;
} action_table[6][CHAR_SET_SIZE];

/*------------------------------------------------------------------*/

void echo();
void echoline();
void cleartoend();
void clearline();
void backspace();
void quote_edit();
void edit_0();
void edit_1();
void edit_2();
void edit_3();
void edit_4();
void edit_5();
void bell();
void insert();

/*------------------------------------------------------------------*/
/*
The following routines are action routines that are executed by the
editor to carry out commands. Each routine has a single character
argument. Each routine is invoked with the character that caused it
to be invoked as its argument.

The argument isn't always useful, but it is included to provide a
consistent interface for the routines.
*/
/*------------------------------------------------------------------*/
/*
Given a specific directory and the starting string of a file name,
find the longest partial file name that starts with the substring.
*/
void
complete_file_name(name, dir)
    char *name;
    char *dir;
{
    DIR *dirp;

    int len, i, nfiles;
    int maxlen;
    int oldlen;

    char oldname[MAXNAMLEN + 1];
    char newname[MAXNAMLEN + 1];

    if ((name != NULL) &&
	((oldlen = strlen(name)) > 0))
    {
	maxlen = oldlen;
	(void) strcpy(oldname, name);
	(void) strcpy(newname, name);
        nfiles = scan_path_dirs (name, dir);

	/* find the longest filename starting with name */

	for (i = 0; i < nfiles; i++)
	{
	    len = strlen(listing[i]);
	    if ((maxlen < len) &&
		    (strncmp(oldname, listing[i], oldlen) == 0))
	        {
		maxlen = len;
		(void) strcpy(newname, listing[i]);
	    }
	}

	/* find the longest common sub string */

	for (i = 0; i < nfiles; i++)
	{
	    len = strlen(listing[i]);
	    if ((len <= maxlen) &&
		    (strncmp(oldname, listing[i], oldlen) == 0))
	    {
		for (;
		    (oldlen < len) &&
		    (strncmp(newname, listing[i], len) != 0);
		    len--);

		maxlen = len;
		newname[maxlen] = '\0';
	    }
	}

	if (strlen(name) != strlen(newname))
	{
	    /* return the extended name */

	    (void) strcpy(name, newname);
	}
	else
	{
	    /* no difference so beep */

	    bell('\0');
	}
    }
}
/*------------------------------------------------------------------*/
/*
Hidden parameters to dirselect. They must be hidden because dirselect
is passed as an argument to scandir.
*/

static char *namep;
static int namelen;

/*------------------------------------------------------------------*/
/*
Passed to scandir. It is used to decide which files to display.
*/

static
dirselect(dp)
    struct dirent *dp;
{
    return (strncmp(dp->d_name, namep, namelen) == 0);
}

/*------------------------------------------------------------------*/
/*
Return all the files in the current directory and all directories listed
in PATH that start with a string passed as name.
*/
int
scan_path_dirs(char *name, char *dir)
{
    struct dirent **temp_dlist;
    char *envpath, *i, finalpath[MAXNAMLEN * 10];
    int curfiles, dfiles = 0, j, k, found;
    
    namelen = strlen (name);
    namep = name;
    
    (void) query_path(0);
    strcpy (finalpath, dir);
    strcat (finalpath, ":");
    if (strcmp (currentdir, dir) == 0) {
       envpath = getenv ("PATH");
       strcat (finalpath, envpath);
    }
    i = strtok (finalpath, ":");
    while (i != NULL)
        {
        curfiles = scandir(i, &temp_dlist, dirselect, alphasort);
        if (curfiles > 0) 
            { 
            for (j = 0; j < curfiles; j++) {
                found = 0;
                for (k = 0; k < dfiles; k++)
                    if (strcmp (listing[k], temp_dlist[j]->d_name) == 0)
                        found = -1;
                if ((!found) && (dfiles < MAX_FILES))
                    strcpy (listing[dfiles++], temp_dlist[j]->d_name);
            }
            free ((char *) temp_dlist);
            }
        i = strtok (NULL, ":");
        }
    qsort (listing, dfiles, sizeof (listing[0]), (void *) strcmp);
    return (dfiles);
}

#ifdef DOUBLE_TAB
/*------------------------------------------------------------------*/
/*
List all the files in a given directory that start with the string
passed as name.
*/
void
list_file_names(dir, name)
    char *dir;
    char *name;
{
    struct dirent **dlist;
    int i;
    int nfiles;
    int colwidth;
    int cols;
    int ncols;
    int nlines;
    int alphasort();

    if (dir == NULL || name == NULL)
    {
	return;
    }

    cols = tty_wincols();

    nfiles = scan_path_dirs(name, dir);

    /* determine the longest file name length */

    colwidth = 8;		/* minimum width */
    for (i = 0; i < nfiles; i++)
    {
	if (strlen (listing[i]) > colwidth)
	{
	    colwidth = strlen (listing[i]);
	}
    }

    colwidth += 3;			/* at least 3 spaces between them */

    /* print the names, sorted vertically per column */
    ncols = cols / colwidth;
    if (ncols == 0)
    {
	/* longest filename is wider than the screen */
	ncols = 1;
    }

    nlines = (nfiles + ncols - 1) / ncols;

    if (nfiles > 0)
    {
	for (i = 0; i < nlines; i++)
	{
	    int j;
	    int l;

	    l = 0;
	    for (j = 0; j < ncols; j++)
	    {
		int m;

		m = l + i;
		if (m >= nfiles)
		{
		    break;
		}

		fputs(listing[m], stdout);

		if (j < (ncols - 1))
		{
		    int k;

		    for (k = strlen (listing[m]); k < colwidth; k++)
		    {
			(void) fputc(' ', stdout);
		    }
		    l += nlines;
		}
	    }
	    fputs(ccr, stdout);
	    fputs(cnl, stdout);
	}
    }
}
#endif /* DOUBLE_TAB
/*------------------------------------------------------------------*/
/*
Assuming that there is a file name under the cursor, return a path
and a file name. If there is no path name return "."
*/
int
get_dir_and_name(dir, name, userend, start, middle, tail)
    char *dir;
    char *name;
    int *userend;
    int *start;
    int *middle;
    int *tail;
{
    int dirlen;

    int newstart;

    int punlen;
    char pun[USER_NAME_SIZE + 1];
    struct passwd *userpwd;

    int i;

    /* set the default path and file name */

    dir[0] = '\0';
    name[0] = '\0';

    /* search for the start of the file name */
    /* start will be left pointing to the first character of the path  */

    for ((*start) = point;
	((0 < (*start)) && (line[(*start) - 1] != ' '));
	(*start)--
	);

    /* search for the end of the file name */
    /* tail will be left pointing at the last character of the path */

    for ((*tail) = point - 1;
	(((*tail) < (length - 1)) && (line[(*tail) + 1] != ' '));
	(*tail)++
	);

    /* search for the middle of the file name */
    /* middle will be left pointing at the first character of the last
     * element of the path */

    for ((*middle) = (*tail) + 1;
	((0 < (*middle)) &&
	    (line[(*middle) - 1] != '/') &&
	    (line[(*middle) - 1] != ' '));
	(*middle)--
	);

    /* get current directory */
    (void) query_path(0);

    newstart = (*start);

    if ((line[newstart] == '.') &&
	    ((newstart + 1) < length) &&
	(line[newstart + 1] == '/'))
    {
	/* if it's "./" use current dir */

	newstart++;
	(void) strcpy(dir, currentdir);
    }
    else if ((line[newstart] == '.') &&
	    ((newstart + 1) < length) &&
	    (line[newstart + 1] == '.') &&
	    ((newstart + 2) < length) &&
	(line[newstart + 2] == '/'))
    {
	/* if it's "../" strip off one name from currentdir and use that */

	newstart += 2;
	(void) strcpy(dir, currentdir);
	for (i = strlen(dir); (i > 0) && (dir[i] != '/'); i--)
	{
	    /* nothing */
	}
	dir[i] = '\0';
    }
    else if (line[newstart] != '/')
    {
	/* doesn't start with a "/"? use currentdir */

	(void) strcpy(dir, currentdir);
///	(void) strcat(dir, "/");
    }

    /* add on the rest of the path */

    dirlen = strlen(dir);
    for (i = 0; i < ((*middle) - newstart); i++)
    {
	dir[dirlen + i] = line[newstart + i];
    }
    dir[dirlen + i] = '\0';

    /* copy file name from line to name */

    for (i = 0; i < ((*tail) - (*middle) + 1); i++)
    {
	name[i] = line[(*middle) + i];
    }
    name[i] = '\0';

    return (TRUE);
}
/*------------------------------------------------------------------*/
/*
Perform file name completion in much the same style as csh.
*/
/*ARGSUSED*/
void
complete_file(ch)
    unsigned char ch;
{
    char dir[10 * (MAXNAMLEN + 1)];
    char name[MAXNAMLEN + 1];

    int userend;
    int start;
    int middle;
    int tail;

    char newline[BUFFER_SIZE];
    int newlength;
    int newpoint;

    int userlen;
    int len;

    int i;

#ifdef DOUBLE_TAB
/* Hops dbl press does show_files() */
    if (last_char == ch )
    {
        show_files(ch);
        last_char = ' ';
        return;
    }
    last_char = ch;
#endif /* DOUBLE_TAB */

    /* get the path and file name in the line */

    if (get_dir_and_name(dir,
	    name,
	    &userend,
	    &start,
	    &middle,
	    &tail))
    {
	/* complete the file name if possible */

	complete_file_name(name, dir);

	/* create a new line */

	/* start with the line prefix */

	(void) strncpy(newline, line, start);
	newline[start] = '\0';

	/* put in the existing path */

	len = strlen(newline);
	(void) strncat(newline, &line[start], middle - start);
	newline[len + (middle - start)] = '\0';

	/* stick in the new file name */

	(void) strcat(newline, name);
	newpoint = strlen(newline);

	/* finish with the line postfix */

	(void) strncat(newline, &line[tail + 1], (length - tail - 1));
	newlength = strlen(newline);

	/* display the new line */

	clearline('\0');

	point = newpoint;
	length = newlength;
	(void) strncpy(line, newline, newlength);

	echoline(line, length);

	for (i = point; i < length; i++)
	{
	    backspace(line[i]);
	}
    }
    else
    {
	bell('\0');
    }
}
#ifdef DOUBLE_TAB
/*------------------------------------------------------------------*/
/*
List the names of files that start with the directory path and
file name under the cursor.
*/
/*ARGSUSED*/
void
show_files(ch)
    unsigned char ch;
{
    void retype_line();

    char dir[10 * (MAXNAMLEN + 1)];
    char name[MAXNAMLEN + 1];

    int userend, start, middle, tail;

    if (get_dir_and_name(dir,
	    name,
	    &userend,
	    &start,
	    &middle,
	    &tail))
    {

       if (strlen (name) == 0)
            bell('\0');
        else {
	    fputs(ccr, stdout);
	    fputs(cnl, stdout);

	    fputs(ccr, stdout);
	    fputs(cnl, stdout);

	    list_file_names(dir, name);

	    retype_line('\0');
	}
    }
    else
    {
	bell('\0');
    }
}
#endif /* DOUBLE_TAB */
/*------------------------------------------------------------------*/
/*
Make the gross assumption that the program we are talking to is
a shell and send "pwd\n" to it. Whatever comes back is saved as the
value of currentdir.
*/
/*ARGSUSED*/
void
query_path(ch)
    unsigned char ch;
{
    static char command[] = "pwd\n";
    char buffer[BUFFER_SIZE];
    int readfd;
    struct timeval timeout;
    int status;
    int cc;
    int i;

    /* send the command to the shell, we hope. */

    write_line(command, strlen(command));

    /* read the directory path back */

    readfd = 1 << master_pty;
    timeout.tv_sec = 2;
    timeout.tv_usec = 0;

    do
    {
	status = select(32,
	    (fd_set *) & readfd,
	    (fd_set *) NULL,
	    (fd_set *) NULL,
	    &timeout);

	if (0 < status)
	{
	    cc = read(master_pty, currentdir, sizeof(currentdir));

	    /* strip off trailing control chars and blanks */

	    for (i = cc - 1; (currentdir[i] <= ' ') && (i > 0); i--)
	    {
		currentdir[i] = '\0';
	    }

	    /* read the prompt so it can be ignored */

	    readfd = 1 << master_pty;
	    timeout.tv_sec = 2;
	    timeout.tv_usec = 0;

	    do
	    {
		status = select(32,
		    (fd_set *) & readfd,
		    (fd_set *) NULL,
		    (fd_set *) NULL,
		    &timeout);

		if (0 < status)
		{
		    cc = read(master_pty, buffer, sizeof(buffer));
		}
		else if ((-1 == status) && windowchanged)
		{
		    windowchanged = FALSE;
		}
	    } while (status == -1);
	}
	else if ((-1 == status) && windowchanged)
	{
	    windowchanged = FALSE;
	}
    } while (status == -1);
    return;
}
/*------------------------------------------------------------------*/
/*
Ring the bell on the terminal.
*/
/*ARGSUSED*/
void
bell(ch)
    unsigned char ch;
{
    fputs(cbl, stdout);
}
/*------------------------------------------------------------------*/
/*
Pass characters to the slave. Don't mess with them at all.
*/
void
pass(ch)
    unsigned char ch;
{
    (void) write(master_pty, &ch, 1);
}
/*------------------------------------------------------------------*/
/*
Insert a character at point in the line buffer. While we are at it
update the display to show the insertion.
*/
void
insert(ch)
    unsigned char ch;
{
    int i;

#ifdef DOUBLE_TAB
    last_char = ' ';
#endif /* DOUBLE_TAB */
    if (length < (BUFFER_SIZE - 2))
    {

	/* display the character */

	echo(ch);

	/* redisplay the rest of the line */

	echoline(&line[point], (length - point));

	/* move the characters in the line buffer */
	/* and put the cursor back at point */

	for (i = length; i > point; i--)
	{
	    line[i] = line[i - 1];
	    backspace(line[i]);
	}

	/* add the character to the line buffer */
	/* and increment point and length */

	line[point] = ch;
	length++;
	point++;
    }
    else
    {
	bell('\0');
    }
}
/*------------------------------------------------------------------*/
/*
Delete a character at point in the line buffer. While we are at it
update the display to reflect the deletion.
*/
/*ARGSUSED*/
void
delete_char_under(ch)
    unsigned char ch;
{
    int i;

/* Hops - ^d at empty line */
    if ( length == 0 && ch == CD )
    {
        pass(ch);
        return;
    }    

    if (point < length)
    {

	/* clear to the end of the line */

	cleartoend();

	/* retype the rest of the line */

	echoline(&line[point + 1], (length - point - 1));

	/* build the new line */

	for (i = point + 1; i < length; i++)
	{
	    line[i - 1] = line[i];
	    backspace(line[i]);
	}

	length--;

	if (point > length)
	{
	    point = length;
	}
    }

}
/*------------------------------------------------------------------*/
/*
Delete the character to the left of point in the line buffer. While we
are at it update the display to reflect the deletion.
*/
/*ARGSUSED*/
void
delete_char(ch)
    unsigned char ch;
{
    int i;

    if (point > 0)
    {
	/* move the cursor left one character */

	backspace(line[point - 1]);

	/* clear to the end of the line */

	cleartoend();

	/* retype the rest of the line */

	echoline(&line[point], (length - point));

	/* build the new line */

	for (i = point; i < length; i++)
	{
	    line[i - 1] = line[i];
	    backspace(line[i]);
	}

	length--;
	point--;
    }

}
/*------------------------------------------------------------------*/
/*
Bind the edit vector to quote_edit so that the next character
will be placed in the line buffer.
*/
/*ARGSUSED*/
void
quote(ch)
    unsigned char ch;
{
    edit = quote_edit;
}
/*------------------------------------------------------------------*/
/*
The next character will select an action from action_table[1]
*/
/*ARGSUSED*/
void
escape_1(ch)
    unsigned char ch;
{
    edit = edit_1;
}
/*------------------------------------------------------------------*/
/*
The next character will select an action from action_table[2]
*/
/*ARGSUSED*/
void
escape_2(ch)
    unsigned char ch;
{
    edit = edit_2;
}
/*------------------------------------------------------------------*/
/*
The next character will select an action from action_table[3]
*/
/*ARGSUSED*/
void
escape_3(ch)
    unsigned char ch;
{
    edit = edit_3;
}
/*------------------------------------------------------------------*/
/*
The next character will select an action from action_table[4]
*/
/*ARGSUSED*/
void
escape_4(ch)
    unsigned char ch;
{
    edit = edit_4;
}
/*------------------------------------------------------------------*/
/*
The next character will select an action from action_table[5]
*/
/*ARGSUSED*/
void
escape_5(ch)
    unsigned char ch;
{
    edit = edit_5;
}
/*------------------------------------------------------------------*/
/*
Move the cursor to the start of the line.
*/
/*ARGSUSED*/
void
start_of_line(ch)
    unsigned char ch;
{
    int i;

    if (length > 0)
    {
	for (i = 0; i < point; i++)
	{
	    backspace(line[i]);
	}
	point = 0;
    }
}
/*------------------------------------------------------------------*/
/*
Move the cursor one character to the left.
*/
/*ARGSUSED*/
void
backward_char(ch)
    unsigned char ch;
{
    if ((length > 0) && (point > 0))
    {
	backspace(line[point - 1]);
	point--;
    }
}
/*------------------------------------------------------------------*/
/*
Move the cursor to the right of the last character on the line.
*/
/*ARGSUSED*/
void
end_of_line(ch)
    unsigned char ch;
{
    if ((length > 0) && (point < length))
    {
	echoline(&line[point], (length - point));
	point = length;
    }
}
/*------------------------------------------------------------------*/
/*
Move the cursor one character to the right.
*/
/*ARGSUSED*/
void
forward_char(ch)
    unsigned char ch;
{
    if ((length > 0) && (point < length))
    {
	echo(line[point]);
	point++;
    }
}
/*------------------------------------------------------------------*/
/*
Add a line to the history buffer and pass it to the child process
as input.
*/
/*ARGSUSED*/
void
add_to_history(ch)
    unsigned char ch;
{
    /* Put the line in the history buffer. Make here point to the current
     * line. And increment head to point to the next history slot. */

    /* If the current line is identical to the current history line, don't
     * add it. */

    /* don't save blank lines */

    int prev;

    if ((head - 1) < 0)
    {
	prev = HISTORY_SIZE - 1;
    }
    else
    {
	prev = head - 1;
    }

    if ((length != 0) &&
	((length != hist[prev].length) ||
	    (strncmp(hist[prev].line, line, length) != 0)))
    {
	/* set the length of the entry */

	hist[head].length = length;

	/* make sure there is enough storage for the new line */

	if (hist[head].line == NULL)
	{
	    if ((hist[head].line = (char *) malloc((unsigned) length)) == NULL)
	    {
		perror("ile - malloc failed");
	    }
	}
	else
	{
	    if ((hist[head].line =
		    (char *) realloc(hist[head].line, (unsigned) length))
		== NULL)
	    {
		perror("ile - realloc failed");
	    }
	}

	(void) strncpy(hist[head].line, line, length);

	head = (head + 1) % HISTORY_SIZE;

	if (hist[head].line != NULL)
	{
	    free(hist[head].line);
	    hist[head].length = 0;
	    hist[head].line = NULL;
	}
    }

    /* reset here */

    here = head;

    /* Echo a carriage return or a newline as a cr-nl sequence. Then send the
     * line to the child process. Finally, clear the buffer for reuse. */

    fputs(ccr, stdout);
    fputs(cnl, stdout);

    line[length] = nl;
    length++;

    write_line(line, length);

    point = 0;
    length = 0;

}
/*------------------------------------------------------------------*/
/*
Erase the entire line.
*/
/*ARGSUSED*/
void
erase_line(ch)
    unsigned char ch;
{
    /* remove any text from the display */

    clearline(ch);

    /* nothing in the line buffer */

    point = 0;
    length = 0;

    /* reset here */

    here = head;

}
/*------------------------------------------------------------------*/
/*
Erase from the current cursor position to the end of the line.
*/
/*ARGSUSED*/
void
erase_to_end_of_line(ch)
    unsigned char ch;
{
    if ((length > 0) && (point < length))
    {
	cleartoend();
	length = point;
    }

}
/*------------------------------------------------------------------*/
/*
Retype the current contents of the edit buffer.
*/
/*ARGSUSED*/
void
retype_line(ch)
    unsigned char ch;
{
    int i, readfd, status, cc;
    struct timeval timeout;

    fputs(ccr, stdout);
    fputs(cnl, stdout);

    fputs (getenv ("PS1"), stdout);

    echoline(line, length);

    for (i = point; i < length; i++)
    {
	backspace(line[i]);
    }
}
/*------------------------------------------------------------------*/
/*
Go to the the next entry in the history buffer and display it.
If we are past the last history entry, then beep.
*/
void
forward_history(ch)
    unsigned char ch;
{
    if (here != head)
    {
	clearline(ch);

	here = (here + 1) % HISTORY_SIZE;
	length = hist[here].length;
	point = length;

	(void) strncpy(line, hist[here].line, length);
	echoline(line, length);
    }
    else
    {
	bell('\0');
    }
}
/*------------------------------------------------------------------*/
/*
Go back one entry in the history buffer and display it. If we are
already at the last entry, then beep.
*/
void
backward_history(ch)
    unsigned char ch;
{
    int prev;

    prev = here - 1;

    if (prev < 0)
    {
	prev = HISTORY_SIZE - 1;
    }

    if (hist[prev].line != NULL)
    {
	clearline(ch);

	here = prev;
	length = hist[here].length;
	point = length;

	(void) strncpy(line, hist[here].line, length);
	echoline(line, length);
    }
    else
    {
	bell('\0');
    }
}
/*------------------------------------------------------------------*/
/*
The following routines are utility routines used by the editing
routines.
*/
/*------------------------------------------------------------------*/
/*
Clear to the end of the current input line.
*/
void
cleartoend()
{
    /* send the clear character */

     fputs(cce, stdout);

    /* send somes nulls for padding */

    fputs("\0\0\0\0", stdout);
    fflush(stdout);
}
/*------------------------------------------------------------------*/
/*
Clear the input line. Backspace to the start of the line. Then clear
to the end of the line.
*/
/*ARGSUSED*/
void
clearline(ch)
    unsigned char ch;
{
    int i;

    for (i = 0; i < point; i++)
    {
	backspace(line[i]);
    }

    cleartoend();
}
/*------------------------------------------------------------------*/
/*
Echo a character. Not all characters are created equal. Control characters
are echoed in ^X form. So they take up two character positions instead of
the normal 1 character position.
*/
void
echo(ch)
    unsigned char ch;
{
    /* how should we echo the char? */

    if (ch < ' ')
    {
	(void) fputc('^', stdout);
	(void) fputc('@' + ch, stdout);
    }
    else
    {
	(void) fputc(ch, stdout);
    }
    (void) fflush(stdout);
    
}
/*------------------------------------------------------------------*/
/*
Echo a line. Print a whole line with control characters printed in
^X form.
*/
void
echoline(line, length)
    char *line;
    int length;
{
    int i;

    for (i = 0; i < length; i++)
    {
	echo(*line++);
    }

}
/*------------------------------------------------------------------*/
/*
Backspace over a character. Generate enough bs characters to backspace
over any character.
*/
void
backspace(ch)
    unsigned char ch;
{
    if (ch < ' ')
    {

	fputs(cle, stdout);
	fputs(cle, stdout);
    }
    else
    {
	fputs(cle, stdout);
    }
}
/*------------------------------------------------------------------*/
/*
Add any character to the line buffer.
*/
void
quote_edit(ch)
    unsigned char ch;
{
    insert(ch);

    edit = edit_0;
}
/*------------------------------------------------------------------*/
/*
Given a character and an action table number either execute the
action or pass the string to (*edit)(ch)
*/
void
dispatch(table, ch)
    int table;
    unsigned char ch;
{
    char *cptr;

    switch (action_table[table][ch].flag)
    {
    case is_action:

	(*(action_table[table][ch].aors.action)) (ch);
	break;

    case is_string:

	cptr = action_table[table][ch].aors.string;
	while ((*cptr) != '\0')
	{
	    (*edit) (*cptr);
	    cptr++;
	}

	break;
    }
}
/*------------------------------------------------------------------*/
/*
Select an action from action_table[5] and execute it.
*/
void
edit_5(ch)
    unsigned char ch;
{
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */

    edit = edit_0;
    dispatch(5, ch);
    (void) fflush(stdout);
}
/*------------------------------------------------------------------*/
/*
Select an action from action_table[4] and execute it.
*/
void
edit_4(ch)
    unsigned char ch;
{
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */

    edit = edit_0;
    dispatch(4, ch);
    (void) fflush(stdout);
}
/*------------------------------------------------------------------*/
/*
Select an action from action_table[3] and execute it.
*/
void
edit_3(ch)
    unsigned char ch;
{
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */

    edit = edit_0;
    dispatch(3, ch);
    (void) fflush(stdout);
}
/*------------------------------------------------------------------*/
/*
Select an action from action_table[2] and execute it.
*/
void
edit_2(ch)
    unsigned char ch;
{
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */

    edit = edit_0;
    dispatch(2, ch);
    (void) fflush(stdout);
}
/*------------------------------------------------------------------*/
/*
Select an action from action_table[1] and execute it.
*/
void
edit_1(ch)
    unsigned char ch;
{
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */

    edit = edit_0;
    dispatch(1, ch);
    (void) fflush(stdout);
}
/*------------------------------------------------------------------*/
/*
Select an action from action_table[0] and execute it.
*/
void
edit_0(ch)
    unsigned char ch;
{
    dispatch(0, ch);
    (void) fflush(stdout);
}
/*------------------------------------------------------------------*/
/*
Input line editor.

Initialize the world. Then loop forever using select to wait for
characters to be available from either stdin or from master_pty.
When characters are available, pass them on after doing any needed
editing.
*/
void
ile()
{
    /* general purpose integer variable */

    int i;

    /* arguments for read and write calls */

    char buffer[BUFFER_SIZE];
    int cc;

    /* Arguments for select call */

    int nfds;
    int width;
    int readfds;

    /* what to do if the child or parent dies */

    (void) signal(SIGCHLD, (void (*)()) clean_up);
    (void) signal(SIGSEGV, (void (*)()) clean_up);
    (void) signal(SIGBUS,  (void (*)()) clean_up);
    (void) signal(SIGTERM, (void (*)()) clean_up);
    (void) signal(SIGHUP,  (void (*)()) clean_up);
    (void) signal(SIGINT,  (void (*)()) clean_up);
    (void) signal(SIGQUIT, (void (*)()) clean_up);

    /* what to do it the window changes size */

    (void) signal(SIGWINCH, change_window);

    /* replicate current tty state onto slave */
    tty_getmode(READ);
    tty_replic(READ, slave_tty) ;

    (void) tty_raw(READ);

    windowchanged = FALSE;

    /* "login" the user */

    (void) setutmp(slave_tty, TRUE);

    /* set raw mode on tty */
    (void) tty_raw(READ);

    kill (0, SIGWINCH); /* set the initial window size */

    /* get descriptor table size */
    width = getdtablesize();
    if (width > 32)
    {
	width = 32;
    }

    /* set initial edit function */

    edit = edit_0;

    /* initialize line buffer */

    point = 0;
    length = 0;

    /* initialize history buffer */

    head = 0;
    here = 0;

    for (i = 0; i < HISTORY_SIZE; i++)
    {
	hist[i].length = 0;
	hist[i].line = NULL;
    }

    for (;;)
    {
	readfds = (1 << READ) | (1 << master_pty);

	/* wait for input from stdin or master_pty */

	nfds = select(width,
	    (fd_set *) & readfds,
	    (fd_set *) NULL,
	    (fd_set *) NULL,
	    (struct timeval *) NULL);

	if (nfds == -1)		/* an exception has occured */
	{
	    if (windowchanged)
	    {
		/* nothing serious, the window changed size */

		windowchanged = FALSE;
	    }
	    else
	    {
		perror("ile - select exception");
		clean_up();
	    }
	}
	else if ((nfds > 0) && (readfds != 0))	/* something to read */
	{
	    if ((readfds & (1 << master_pty)) != 0)
	    {
		/* read the pending characters. */

		cc = read(master_pty, buffer, BUFFER_SIZE);

		/* display the characters. */

		(void) write(WRITE, buffer, cc);
	    }

	    if ((readfds & (1 << READ)) != 0)
	    {
		/* read the pending characters. */

		cc = read(READ, buffer, BUFFER_SIZE);

		/* if the slave is not echoing chars
		 *  should not mess with its input characters
		 */
                if ( tty_silentmode(slave_tty) )
		{
		    edit = pass;
		}
		else if (edit == (void (*) ()) pass)
		{
		    edit = edit_0;
		}

		/* decide what to do with the characters. */

		for (i = 0; i < cc; i++)
		{
/* printf("%2x ", buffer[i]); */
		    (*edit) (CHAR_MASK & buffer[i]);
		}
	    }

	}
    }

}
/*------------------------------------------------------------------*/
/*
The child process.

Make the pty the processes controling terminal. Bind the pty to
stdin, stdout, and stderr. Then exec the users program.
*/
void
child(argv)
    char *argv[];
{
    /* shell name pointers */

    char *shellname;
    char *shellpath;
    char *dashshellname;

    /* close all file descriptors */

    (void) close(READ);
    (void) close(WRITE);
    (void) close(ERROR);
    (void) close(slave_tty);

    /* get rid of controlling terminal */
    tty_NoCtty();

    /* open the tty again */
    /* this makes the pty the controlling terminal */

    if ((slave_tty = open(ttydev, O_RDWR)) == -1)
    {
	perror("ile - reopening slave tty");
    }

    /* slave_tty is now stdin */

    /* bind slave_tty to stdout */

    (void) dup2(slave_tty, WRITE);

    /* bind slave_tty to stderr */

    (void) dup2(slave_tty, ERROR);

    /* close master_pty descriptor */

    (void) close(master_pty);

    /* Fire up application program. If no program name is given then fire up
     * users favorite shell. */

    /* get the name of the users shell. default to /bin/csh */

    if ((shellpath = getenv("SHELL")) == NULL ||
	(*shellpath == '\0'))
    {
	shellpath = "/bin/csh";
    }

    /* get just the name */

    if ((shellname = strrchr(shellpath, '/')) != NULL)
    {
	shellname += sizeof(char);
    }

    /* if the current argv[0] starts with -, then the new argv[0] */
    /* should start with - */

    if (*(argv[0]) == '-')
    {
	dashshellname = (char *) malloc((unsigned) strlen(shellname) + 2);
	(void) strcpy(dashshellname, "-");
	(void) strcat(dashshellname, shellpath);
    }
    else
    {
	dashshellname = shellname;
    }

#ifdef SIGTSTP
        /* Prevent non-job-control programs from dying on SIGTSTP */
                signal(SIGTSTP, SIG_IGN);
#endif

    /* execute the shell or the specified program */

    if (argv[1] == NULL)
    {
	/* execute default shell */

	execlp(shellpath, dashshellname, 0);
    }
    else if (*argv[1] == '-')
    {
	/* there is an initialization file */

	if (argv[2] == NULL)
	{
	    /* execute default shell */

	    execlp(shellpath, dashshellname, 0);
	}
	else
	{
	    /* execute specified program */

	    execvp(argv[2], &argv[2]);
	}
    }
    else
    {
	/* execute specified program */

	execvp(argv[1], &argv[1]);
    }

    /* this executes if exec fails */

    perror("ile exec failed");
    exit(1);
    /* NOTREACHED */
}
/*------------------------------------------------------------------*/
/*
Set up default key bindings and delimeters.
*/
void
default_bindings()
{
    int i;

    /* clear delimiter vector and the action table */

    for (i = 0; i < CHAR_SET_SIZE; i++)
    {
	delimit[i] = FALSE;

	action_table[0][i].aors.action = insert;
	action_table[1][i].aors.action = bell;
	action_table[2][i].aors.action = bell;
	action_table[3][i].aors.action = bell;
	action_table[4][i].aors.action = bell;
	action_table[5][i].aors.action = bell;

	action_table[0][i].flag = is_action;
	action_table[1][i].flag = is_action;
	action_table[2][i].flag = is_action;
	action_table[3][i].flag = is_action;
	action_table[4][i].flag = is_action;
	action_table[5][i].flag = is_action;
    }

    /* default delimiters */

    delimit[' '] = TRUE;	/* blank */
    delimit['/'] = TRUE;	/* slash */
    delimit['.'] = TRUE;	/* dot */
    delimit['-'] = TRUE;	/* dash */

    /* default action_table[0] */

    action_table[0][CA].aors.action = start_of_line;
    action_table[0][CB].aors.action = backward_char;
    action_table[0][CE].aors.action = end_of_line;
    action_table[0][CF].aors.action = forward_char;
    action_table[0][CI].aors.action = complete_file;
    action_table[0][CK].aors.action = erase_to_end_of_line;
    action_table[0][CU].aors.action = erase_line;
    action_table[0][CL].aors.action = retype_line;
    action_table[0][CN].aors.action = forward_history;
    action_table[0][CP].aors.action = backward_history;
    action_table[0][CV].aors.action = quote;
    action_table[0][del].aors.action = delete_char;
    action_table[0][esc].aors.action = escape_1;
    action_table[0][cr].aors.action = add_to_history;
    action_table[0][nl].aors.action = add_to_history;
    action_table[0][CX].aors.action = delete_char_under;

    action_table[0][CC].aors.action = pass;
    action_table[0][CD].aors.action = delete_char_under /* pass */;
    action_table[0][CQ].aors.action = pass;
    action_table[0][CS].aors.action = pass;
    action_table[0][CZ].aors.action = pass;

    /* default action_table[1] ^[ c */

    action_table[1]['['].aors.action = escape_2;
    action_table[1][esc].aors.action = complete_file;

    /* default action_table[2] ^[ [ c */

    action_table[2]['A'].aors.action = backward_history;
    action_table[2]['B'].aors.action = forward_history;
    action_table[2]['C'].aors.action = forward_char;
    action_table[2]['D'].aors.action = backward_char;
    action_table[2]['3'].aors.action = escape_3;
    action_table[2]['4'].aors.action = escape_4;
    action_table[2]['1'].aors.action = escape_5;
    
    /* default action_table[3] ^[ [ 3 c */
    
    action_table[3]['~'].aors.action = delete_char_under;
    
    /* default action_table[4] ^[ [ 4 c */
    
    action_table[4]['~'].aors.action = end_of_line;
    
    /* default action_table[5] ^[ [ 1 c */
    
    action_table[5]['~'].aors.action = start_of_line;
    

}
/*------------------------------------------------------------------*/
/*
Return a character or EOF. This routine reads characters from input
and converts them into a character using the following rules.

The character may be a single character, a control
character indicated by ^x, an octal number starting with \, or an
escaped character indictated by \x.
*/
int
scan_char(input)
    FILE *input;
{
    int ch;
    int value;

    ch = fgetc(input);
    switch (ch)
    {
    case '^':

	/* it is a control character */

	for (ch = fgetc(input); '@' <= ch; ch = ch - '@');

	break;

    case '\\':

	/* octal or an escaped character? */

	ch = fgetc(input);
	if (('0' <= ch) && (ch <= '7'))
	{

	    /* its an octal number */

	    value = 0;
	    while (('0' <= ch) && (ch <= '7'))
	    {
		value = (value * 8) + (ch - '0');
		ch = fgetc(input);
	    }
	    (void) ungetc(ch, input);

	    ch = value & 0377;	/* make sure it is in range */
	}
	else
	{
	    /* its an escaped character */

	    ch = fgetc(input);
	}

	break;

    case '\n':

	/* the real end of the line */

	ch = EOL;

	break;

    default:

	/* it is just itself */

	break;
    }

    return (ch);

}
/*------------------------------------------------------------------*/
/*
*/
/*ARGSUSED*/
main(argc, argv)
    int argc;
    char *argv[];
{

    /* Child process id */

    int childpid;

    /* identify yourself */

    if (setvbuf(stdout, NULL, _IONBF, 0) < 0 )
        perror("setbuf");
        
    (void) fprintf (stdout, "ile rev 2.7freesco\r\n");
    
    /* create the tty/pty pair */

    getpty(&master_pty, &slave_tty);

    /* get control sequences from termcap */

    get_termcap();

    /* initialize the default bindings */

    default_bindings();

    /* create the child process */

    childpid = fork();

    switch (childpid)
    {
    case 0:			/* child process */

	child(argv);
	break;

    case -1:			/* fork failed */

	perror("ile - fork failed");
	exit(1);
	/* NOTREACHED */

    default:			/* parent process */
	ile();
	break;
    }

}

